/*! \file   blcluster.h
    \brief  Include file for the class blcluster
    \author M. Bebendorf
    \date   4/24/2004                            */

#ifndef BLCLUSTER_H
#define BLCLUSTER_H

#include <stdlib.h>
#include <assert.h>
#include "mblock.h"
#include "cluster.h"

/*! \brief generate block cluster tree, returns a pointer to this tree
    \param cl1   pointer to a block cluster tree for the degrees of freedoms of
                 the rows (input)
    \param cl2   pointer to a block cluster tree for the degrees of freedoms of
                 the columns (input)
    \param eta    the cluster parameter eta (input)
    \param nblcks the number of generated blcks (output)

    \par Description:
    This function generates a block cluster tree by subdividing each block
    into four subblocks. The tree is terminated with blocks that are either
    admissible or one of the dimensions is smaller than bmin. In this case the
    first son of such a cluster will point to NULL. */

extern blcluster *genblcltree(cluster* cl1, cluster* cl2, 
			      double eta, unsigned& nblcks);
extern blcluster *genblcltreeND(cluster* cl1, cluster* cl2, 
			      double eta, unsigned& nblcks);

/*! \brief generate a symmetric block cluster tree, returns a pointer to
           this tree

    \param cl   pointer to a block cluster tree for the degrees of freedoms of
                 the rows (input)
    \param eta    the cluster parameter eta (input)
    \param nblcks the number of generated blcks (output)

    \par Description:
    This function generates a block cluster tree by subdividing each block
    into four subblocks. The tree is terminated with blocks that are either
    admissible or one of the dimensions is smaller than bmin. In this case the
    first son of such a cluster will point to NULL.                     
    In contrast to genblcltree this function sets the third son of each
    diagonal block to NULL. */

extern blcluster *gensymblcltree(cluster* cl, double eta, unsigned& nblcks);
extern blcluster *gensymblcltreeND(cluster* cl, double eta, unsigned& nblcks);

extern void genpart_from_tree(blcluster*, unsigned, blcluster**&);
extern void genpart(cluster*, cluster*, unsigned, double, unsigned&,
		    blcluster**&);

extern blcluster *genblcltree_augm(unsigned N, unsigned k, cluster* cl1,
				   cluster* cl2, double eta, unsigned& nblcks);
extern blcluster *gensymblcltree_augm(unsigned N, unsigned k, cluster* cl,
				      double eta, unsigned& nblcks);

//! in this class the block structure of the H-matrix is stored
struct blcluster
{
  //! the row index of the first entry of this block within the H-matrix
  unsigned b1;

  //! the column index of the first entry of this block within the H-matrix
  unsigned b2;

  //! the number of rows of this block
  unsigned n1;

  //! the number of columns of this block
  unsigned n2;

  unsigned icom;        // index which is closest to com of cl1

  //! the type of this block (0 - diagonal, -1 - sub-, +1 superdiagonal block)
  char pos;

  //! points to the upper left subblock, son1==0 if the whole block is a leaf
  blcluster *son1;

  //! points to the upper right subblock, son2==0 if son2 does not exist
  blcluster *son2;

  //! points to the lower left subblock, son3==0 if son3 does not exist
  blcluster *son3;

  //! points to the lower riht subblock
  blcluster *son4;

  // test ***
  //! the row cluster and the column cluster
  cluster *rcl, *ccl;           // brauchen nun kein n1,n2 und b1,b2 mehr!!!
  // ***

  //! the index of this block, only valid if this block is a leaf (son1==0)
  unsigned idx;

  //! is this block admissible ? 
  bool adm;

  //! empty constructor
  blcluster() : rcl(NULL), ccl(NULL) { son1 = son2 = son3 = son4 = NULL; }

  //! constructor, initializes this blcluster with bl
  blcluster(blcluster* bl) : n1(bl->n1), b1(bl->b1), n2(bl->n2), b2(bl->b2),
			     idx(bl->idx), adm(bl->adm), pos(bl->pos),
			     icom(bl->icom), rcl(bl->rcl), ccl(bl->ccl) {
    if (bl->isleaf()) son1 = NULL;
    else {
      son1 = new blcluster(bl->son1);
      son2 = new blcluster(bl->son2);
      if (bl->son3) son3 = new blcluster(bl->son3);
      else son3 = NULL;
      son4 = new blcluster(bl->son4);
    }
  }

  //! destructor
  ~blcluster() {
    if (son1) {
      delete son1;
      son1 = NULL;
      if (son2) delete son2;
      if (son3) delete son3;
      delete son4;
    }
  }

  //! is this a leaf ?
  bool isleaf() const { return (son1==NULL); }

  //! is this a diagonal block ?
  bool isdbl() const { return (pos==0); }

  //! is this block in upper triangular part ?
  bool is_in_utr() const { return (pos>=0); }

  //! is this block in lower triangular part ?
  bool is_in_ltr() const { return (pos<=0); }

  //! is this block admissible ?
  bool isadm() const { return adm; }

  //! is the corresponding block in A stored as a low-rank matrix ?
  bool islwrank(mblock<double>** const A) const
  { return A[idx]->islwrank(); }
  bool islwrank(mblock<float>** const A) const
  { return A[idx]->islwrank(); }
  bool islwrank(mblock<scomp>** const A) const
  { return A[idx]->islwrank(); }
  bool islwrank(mblock<dcomp>** const A) const
  { return A[idx]->islwrank(); }
  

  //! is the corresponding block in A stored as a dense matrix ?
  bool isdense(mblock<double>** const A) const { return A[idx]->isdense(); }
  bool isdense(mblock<float>** const A) const { return A[idx]->isdense(); }
  bool isdense(mblock<scomp>** const A) const { return A[idx]->isdense(); }
  bool isdense(mblock<dcomp>** const A) const { return A[idx]->isdense(); }

  //! is the corresponding block in A stored as a lower triangular matrix ?
  bool isltr(mblock<double>** const A) const
  { return A[idx]->isltr(); }
  bool isltr(mblock<float>** const A) const
  { return A[idx]->isltr(); }
  bool isltr(mblock<scomp>** const A) const
  { return A[idx]->isltr(); }
  bool isltr(mblock<dcomp>** const A) const
  { return A[idx]->isltr(); }

  //! is the corresponding block in A stored as a upper triangular matrix ?
  bool isutr(mblock<double>** const A) const
  { return A[idx]->isutr(); }
  bool isutr(mblock<float>** const A) const
  { return A[idx]->isutr(); }
  bool isutr(mblock<scomp>** const A) const
  { return A[idx]->isutr(); }
  bool isutr(mblock<dcomp>** const A) const
  { return A[idx]->isutr(); }
   
  //! returns the rank of the corresponding block in A
  unsigned rank(mblock<double>** const A) const
  { assert(islwrank(A)); return A[idx]->rank(); }
  unsigned rank(mblock<float>** const A) const
  { assert(islwrank(A)); return A[idx]->rank(); }
  unsigned rank(mblock<scomp>** const A) const
  { assert(islwrank(A)); return A[idx]->rank(); }
  unsigned rank(mblock<dcomp>** const A) const
  { assert(islwrank(A)); return A[idx]->rank(); }

  //! returns the pointer to the data of the corresponding block in A
  double* data(mblock<double>** const A) const { return A[idx]->data; }
  float* data(mblock<float>** const A) const { return A[idx]->data; }
  scomp* data(mblock<scomp>** const A) const { return A[idx]->data; }
  dcomp* data(mblock<dcomp>** const A) const { return A[idx]->data; }
  
  //! returns the size of the block cluster tree starting from this
  unsigned long sizeT() const {
    if (isleaf()) return sizeof(blcluster);
    else {
      unsigned long size = son1->sizeT() + son2->sizeT();
      if (son3) size += son3->sizeT();
      return size + son4->sizeT();
    }
  }

  //! returns the number of leaves in the block cluster tree starting from this
  unsigned long nleaves() const {
    if (isleaf()) return 1;
    else {
      unsigned long nl = son1->nleaves();
      nl += son2->nleaves();
      if (son3) nl += son3->nleaves();
      nl += son4->nleaves();
      return nl;
    }
  }

  //! returns the number of leaves on the block diagonal (i.e. pos==0)
  unsigned long ndbls() const {
    if (isleaf()) return 1;
    else return son1->ndbls() + son4->ndbls();
  }
};

#endif  // BLCLUSTER_H
