/*! \file  cluster.h
    \brief Include file for the class cluster
    \author M. Bebendorf
    \date   3/24/2004
*/


#ifndef CLUSTER_H
#define CLUSTER_H

#include <iostream>
#include "dof.h"

struct blcluster;

//! the basis class for storing clusters of degrees of freedom
class cluster
{
 protected:
  //! squared diameter
  double diam2;

  // index closest to the centroid
  unsigned icom;

 public:
  //! first son of this cluster in the tree, son1==0 if this is a leaf
  cluster *son1;
  //! second son of this cluster in the tree
  cluster *son2;

  // test ***
 public:
  blcluster* dbl; // bl is the diagonal block corresponding to this cluster
                 // = NULL if no such block exists
  // test ***

  //! begining index
  
  unsigned nbeg;

  //! ending index (begining index of next)
  unsigned nend;            

  //! returns the spatial dimension 
  virtual unsigned dim() const = 0;

 public:
  /*!
    \brief Constructor
    \param k begining index of the cluster
    \param l begining index of the next cluster 
  */
  cluster(unsigned k, unsigned l) :
    son1(NULL), son2(NULL), dbl(NULL), nbeg(k), nend(l) {}

    //  cluster() { };

  virtual ~cluster() { if (son1!=NULL) { delete son1; delete son2; } }

  unsigned getnbeg() const { return nbeg; }
  unsigned getnend() const { return nend; }

  cluster* getson1() const { return son1; }
  cluster* getson2() const { return son2; }

  unsigned geticom() const { return icom; }

  bool isnleaf() const { return (son1!=NULL); }

  //! is this cluster admissible with the cluster cl ? 
  virtual bool isadm(double eta, cluster* cl) const = 0;

  unsigned size() const { return nend-nbeg; }

  /*!
    \brief subdivide this cluster recursively, needed to generate the cluster
           tree
    \param dofs the global list of pointers to the degrees of freedom, i.e.,
           dofs[nbeg],...,dofs[nend-1] point to the degrees of freedom
	   contained in this cluster. Upon termination
           dofs is changed and contains the permutation made (input/output)
    \param bmin the minimal size of clusters (input)
    \param nClusters the number of generated clusters; has to been initialized
           with 1 (output)
           
    \verbatim
    cluster3d_pca* clTree = new cluster3d_pca(clindx, 0, nPanels);
    unsigned nClusters = 1;
    clTree->subdivide(clindx, bmin, nClusters);
    \endverbatim
  */
  virtual void subdivide(dof** dofs, unsigned bmin, unsigned& nClusters) = 0;
};

#endif
