#ifndef MBLOCK_H
#define MBLOCK_H

#include <assert.h>
#include <iostream>
#include "blas.h"

template <class typ> class mblock
{
  friend struct blcluster;
  friend void copymbl(mblock<double>*, mblock<float>*);
  friend void copymbl(mblock<dcomp>*, mblock<scomp>*);
  
 protected:
  typ *data;              // if low-rank-repr. UV^T (twice MAX_RANK columns)
  unsigned n1, n2;        // n1 number of rows, n2 number of columns
  unsigned bl_rank;       // the rank of this block

  struct st_info {
    unsigned is_lwrank : 1;      // low rank representation ? else dense
    unsigned is_sym : 1;         // for dense matrices: is symmetric ?
    unsigned is_ltr : 1;         // for dense matrices: is lower triangular ?
    unsigned is_utr : 1;         // for dense matrices: is upper triangular ?
  } info;

  // a low-rank matrix U V^T is stored columnwise : (U,V)
  // a dense matrix is stored column by column
  // a dense symmetric matrix is stored as an upper triangular matrix
  // an upper triangular matrix is stored columnwise (a11,a12,a22,...)
  // a lower triangular matrix is stored columnwise (a11,a21,a31,...)
  //         on the diagonal of it is a permutation P of the rows of L defined
  //         by (Pl)_{L_{ii}} = l_i

  // add low-rank matrix to this low-rank matrix with truncation
  void addtrll(unsigned, typ*, unsigned, typ*, unsigned, double,
	       unsigned);

  // add low-rank matrix to this low-rank matrix with truncation
  // returns remainder
  void addtrll_rmnd(unsigned, typ*, unsigned, typ*, unsigned,
		    double, unsigned, unsigned&, typ*&, typ*&);

  // compute singular values of this low-rank matrix
  void get_svals_lwr(double*) const;
  void get_svals_lwr(float*) const;

  // unify two low-rank matrices
  void unify_cols_lwr_lwr(double, unsigned, const mblock&, const mblock&);

  // **************************************************************************
  // PostScript output for low-rank and dense matrices
  //

  void psout_lwr(std::ostream&, unsigned, unsigned, unsigned, bool,
		 bool) const;
  void psout_dense(std::ostream&, unsigned, unsigned, unsigned, bool) const;

  // **************************************************************************
  // multiplication routines
  //

  // multiply lwr by vector y += d U V^T x
  void multa_lwr_vec(typ d, typ* x, typ* y) const;

  // multiply lwr by vector Y += d U V^T X
  void multa_lwr_dense(typ d, unsigned p, typ* X, unsigned ldX,
		       typ* Y, unsigned ldY) const;

  // multiply transposed lwr by vector y += d V U^T x
  void multa_lwrT_vec(typ d, typ* x, typ* y) const;

  // multiply transposed lwr by vector Y += d V U^T X
  void multa_lwrT_dense(typ d, unsigned p, typ* X, unsigned ldX,
			typ* Y, unsigned ldY) const;

  // multiply dense by vector: y += d A x  (A is a dense matrix) 
  void multa_dense_vec(typ d, typ* x, typ* y) const {
    assert(isdense()); blas::gemva(n1, n2, d, data, x, y);
  }

  // multiply dense by vector: Y += d A X  (A is a dense matrix) 
  void multa_dense_dense(typ d, unsigned p, typ* X, unsigned ldX,
			 typ* Y, unsigned ldY) const {
    assert(isdense()); blas::gemma(n1, n2, p, d, data, n1, X, ldX, Y, ldY);
  }

  // multiply vector by transposed dense: y += A^t x  (A is a dense matrix)
  void multa_denseT_vec(typ d, typ* x, typ* y) const {
    assert(isdense()); blas::gemtva(n1, n2, d, data, x, y);
  }

  // multiply vector by transposed dense: Y += d A^t X  (A is a dense matrix) 
  void multa_denseT_dense(typ d, unsigned p, typ* X, unsigned ldX,
			  typ* Y, unsigned ldY) const {
    assert(isdense()); blas::gemtma(n1, n2, p, d, data, n1, X, ldX, Y, ldY);
  }

  // multiply symmetric dense by vector: y += d A x (A is symm. dense)
  void multa_denseSym_vec(typ d, typ* x, typ* y) const {
    assert(isdense() && issym() && n1==n2); blas::gemsva(n1, d, data, x, y);
  }

  // multiply symmetric dense by vector: Y += d A X (A is symm. dense)
  void multa_denseSym_dense(typ d, unsigned p, typ* X, unsigned ldX,
			   typ* Y, unsigned ldY) const {
    for (unsigned l=0; l<p; ++l) multa_denseSym_vec(d, X+ldX*l, Y+ldY*l); 
  }

  // multiply lower triangular by vector: y += d PL x
  void multa_denseLTr_vec(typ d, typ* x, typ* y) const;

  // multiply lower triangular by vector: y += d PL x
  void multa_denseLTr_dense(typ d, unsigned p, typ* X, unsigned ldX,
			   typ* Y, unsigned ldY) const {
    for (unsigned l=0; l<p; ++l) multa_denseLTr_vec(d, X+ldX*l, Y+ldY*l);
  }

  // multiply upper triangular by vector: y += d A x (A is utr)
  void multa_denseUTr_vec(typ d, typ* x, typ* y) const;

  // multiply upper triangular by vector: Y += d A X (A is utr)
  void multa_denseUTr_dense(typ d, unsigned p, typ* X, unsigned ldX,
			   typ* Y, unsigned ldY) const {
    for (unsigned l=0; l<p; ++l) multa_denseUTr_vec(d, X+ldX*l, Y+ldY*l); 
  }

  // multiply lower triangular by vector: y += d (PL)^T x = d L^T P^{-1} x
  void multa_denseLTrT_vec(typ d, typ* x, typ* y) const;

  // multiply lower triangular by vector: y += d (PL)^T x = d L^T P^{-1} x
  void multa_denseLTrT_dense(typ d, unsigned p, typ* X, unsigned ldX,
			     typ* Y, unsigned ldY) const {
    for (unsigned l=0; l<p; ++l) multa_denseLTrT_vec(d, X+ldX*l, Y+ldY*l);
  }

  // multiply upper triangular by vector: y += d A^T x (A is utr)
  void multa_denseUTrT_vec(typ d, typ* x, typ* y) const;

  // multiply upper triangular by vector: y += d A^T x (A is utr)
  void multa_denseUTrT_dense(typ d, unsigned p, typ* X, unsigned ldX,
			    typ* Y, unsigned ldY) const {
    for (unsigned l=0; l<p; ++l) multa_denseUTrT_vec(d, X+ldX*l, Y+ldY*l);
  }

  // **************************************************************************
  // PUBLIC routines
  //

 public:

  // Konstruktor
  mblock(unsigned m, unsigned n) : n1(m), n2(n) {
    bl_rank = info.is_sym = info.is_ltr = info.is_utr = 0;
    info.is_lwrank = 1; data = NULL;
  }
  
  // Destruktor
  ~mblock() { delete[] data; }

  // return size of this mblock
  unsigned long size() const;

  // **************************************************************************
  // set/read attributes

  bool islwrank() const { return info.is_lwrank; }
  bool isdense() const { return !islwrank(); }
  bool issym() const { assert(isdense()); return info.is_sym; }
  bool isltr() const { assert(isdense()); return info.is_ltr; } 
  bool isutr() const { assert(isdense()); return info.is_utr; }

  unsigned rank() const { assert(islwrank()); return bl_rank; }

  void freedata() {
    delete[] data; data = NULL;
    info.is_lwrank = 1; bl_rank = 0;
  }

  typ* getdata() const { return data; }

  void setrank(const unsigned k) {
    delete[] data;
    bl_rank = k; info.is_lwrank = 1;
    if (k) {
      data = new typ[k*(n1+n2)];
      assert(data!=NULL);
    }
    else data = NULL;
  }

  void setdense() {
    delete[] data;
    info.is_lwrank = info.is_sym = info.is_ltr = info.is_utr = 0;
    data = new typ[n1*n2]; assert(data!=NULL);
  }

  void setdenseSym() {
    assert(n1==n2);
    delete[] data;
    info.is_lwrank = info.is_utr = info.is_ltr = 0;
    info.is_sym = 1;
    data = new typ[n1*(n1+1)/2]; assert(data!=NULL);
  }

  void set_ltr() {
    assert(n1==n2);
    delete[] data;
    info.is_lwrank = info.is_sym = info.is_utr = 0;
    info.is_ltr = 1;
    data = new typ[n1*(n1+1)/2]; assert(data!=NULL);
  }

  void set_utr() {
    assert(n1==n2);
    delete[] data;
    info.is_lwrank = info.is_sym = info.is_ltr = 0;
    info.is_utr = 1;
    data = new typ[n1*(n1+1)/2]; assert(data!=NULL);
  }


  // **************************************************************************
  // Initialization routines
  // 

  // init with 0

  void init0_dense(unsigned m, unsigned n) {
    n1 = m; n2 = n; setdense(); blas::setzero(n1*n2, data);
  }

  void init0_denseSym(unsigned n) {
    n1 = n2 = n; setdenseSym(); blas::setzero(n1*(n1+1)/2, data);
  }

  void init0_utr(unsigned n) {
    n1 = n2 = n; set_utr(); blas::setzero(n1*(n1+1)/2, data);
  }

  void init0_ltr(unsigned n) {
    n1 = n2 = n; set_ltr();
    blas::fill0_ltr(n, data);
  }

  // init with identity
  void initId_dense(unsigned n) {
    n1 = n2 = n;
    setdense();
    blas::fillId(n, data);
  }

  void initId_denseSym(unsigned n) {
    n1 = n2 = n;
    setdenseSym();
    blas::fillId_utr(n, data);
  }

  void initId_ltr(unsigned n) {
    n1 = n2 = n;
    set_ltr();
    blas::fillId_ltr(n, data);
  }
  
  void initId_utr(unsigned n) {
    n1 = n2 = n;
    set_utr();
    blas::fillId_utr(n, data);
  }

  // **************************************************************************
  // Addition routines
  //

  // add dense to dense
  void add_dense_to_dense(typ* const A, const unsigned ldA) {
    assert(ldA>=n1 && isdense());
    for (unsigned j=0; j<n2; ++j) blas::add(n1, A+j*ldA, data+j*n1);
  }

  void add_denseSym_to_denseSym(typ* A) {
    assert(isdense() & issym());
    blas::axpy(n1*(n1+1)/2, D_ONE, A, data);
  }

  void add_dense_to_denseSym(typ* A, const unsigned ldA);

  // add lwr to dense
  void add_lwr_to_dense(unsigned k, typ* U, unsigned ldU,
			typ* V, unsigned ldV)
  {
    assert(isdense());
    if (k>0) blas::gemmta(n1, k, n2, U, ldU, V, ldV, data, n1);
  }

  // add lwr to symm. dense
  void add_lwr_to_denseSym(unsigned mult, unsigned k, typ* U, unsigned ldU,
			   typ* V, unsigned ldV);

  // add lwr to mblock and truncate to min(eps, kgoal)
  void add_lwr(unsigned k, typ* U, unsigned ldU, typ* V, unsigned ldV,
	       double eps, unsigned kgoal);

  // add lwr to mblock and truncate to min(eps, kgoal), returns remainder
  void add_lwr_rmnd(unsigned k, typ* U, unsigned ldU, typ* V,
		    unsigned ldV, double eps, unsigned kgoal,
		    unsigned& kR, typ* &UR, typ* &VR);

  // add dense to mblock and truncate to min(eps, MAX_RANK)
  void add_dense(typ*, unsigned, double, unsigned);

  void unify_cols(double, unsigned, const mblock&, const mblock&);

  void add_unified_rows(double, unsigned, const mblock&, const mblock&);
  void add_unified_rows_lwr_lwr(double, unsigned, const mblock&,
				const mblock&);

  void add_unified(double, unsigned, mblock&, mblock&, mblock&, mblock&);
  void add_unified_rmnd(double, unsigned, mblock&, mblock&, mblock&, mblock&,
			unsigned&, typ*&, typ*&);

  // **************************************************************************
  // mutiply mblock or its transposed by vector

  void multa_mbl_vec(typ d, typ* x, typ* y) const;

  void multa_mbl_dense(typ d, unsigned p, typ* X, unsigned ldX,
		       typ* Y, unsigned ldY) const;

  void multa_mblT_vec(typ d, typ* x, typ* y) const;

  void multa_mblT_dense(typ d, unsigned p, typ* X, unsigned ldX,
			typ* Y, unsigned ldY) const;

  // **************************************************************************
  // Inversion routines
  //

  void invert_dense(typ* dataC) {
    assert(isdense() && n1==n2);
    blas::copy(n1*n1, data, dataC);
    lapack::geinv(n1, dataC);
  }

  void invert_denseSym(typ* dataC) {
    assert(isdense() && issym() && n1==n2);
    blas::copy(n1*(n1+1)/2, data, dataC);
    lapack::geinv_sym(n1, dataC);
  }

  int decomp_Cholesky() {
    assert(isdense() && issym() && n1==n2);
    int inf = lapack::pptrf(n1, data);
    info.is_ltr = info.is_sym = 0;
    info.is_utr = 1;
    return inf;
  }

  int decomp_LU(mblock*, mblock*); // generates LU decomp., destroys data

  // **************************************************************************
  // forward/backward substitution
  //
  
  // solves PL X = B for X, L is unit lower triangular, X is stored in B
  void ltr_solve(unsigned, typ*, unsigned) const;

  // solves (PL)^T X = = B for X, L is unit lower triangular, X is stored in B
  void ltrT_solve(unsigned, typ*, unsigned) const;

  // solves U X = B for X
  // U is upper triangular, X is stored in B
  void utr_solve(unsigned m, typ* B, unsigned ldB) const {
    assert(isdense() && isutr());
    int inf = lapack::utrs(n1, data, m, B, ldB);
    assert(inf==0);
  }

  // solves U^T X = B for X, U is upper triangular, X is stored in B
  void utrT_solve(unsigned m, typ* B, unsigned ldB) const {
    assert(isdense() && isutr());
    int inf = lapack::utrTs(n1, data, m, B, ldB);
    assert(inf==0);
  }

  // solves X U = B for X, U is upper triangular, X is stored in B
  void utr_solve_left(unsigned, typ*, unsigned, typ*, unsigned) const;


  // **************************************************************************
  // Misc routines
  //

  void outp(std::ostream& os) const {
    for (unsigned i=0; i<n1; ++i) {
      for (unsigned j=0; j<n2; ++j) os << data[i+j*n1] << ' ';
      os << std::endl;
    }
  }

  void psout(std::ostream& os, unsigned nmax, unsigned b1, unsigned b2,
	     bool showrank, bool refl) const {
    if (islwrank()) psout_lwr(os, nmax, b1, b2, showrank, refl);
    else psout_dense(os, nmax, b1, b2, refl);
  }

  // copy mblock
  void copy(mblock&);

  // copy transposed of off-diagonal mblocks
  void copyT(mblock&);

  // copy low-rank matrix to this block
  void cpylwr(unsigned k, typ* U, typ* V) {
    setrank(k);
    if (k>0) {
      blas::copy(k*n1, U, data);
      blas::copy(k*n2, V, data+k*n1);
    }
  }

  
  // copy low-rank matrix to this block with recompression
  void cpylwr_cmpr(unsigned k, typ* U, unsigned ldU,
		   typ* V, unsigned ldV,
		   double eps, unsigned rankmax) {
    freedata();
    addtrll(k, U, ldU, V, ldV, eps, rankmax);
  }

  void cpydense(typ* A) {
    setdense();
    blas::copy(n1*n2, A, data);
  }

  void conv_lwr_to_dense();  // internal conversion
  void conv_lwr_to_dense(unsigned, typ*, unsigned, typ*, unsigned);
  void conv_dense_to_lwr(double);   // internal conversion

  void conv_denseSym_dense(typ*, unsigned) const;
  void conv_lwr_to_dense(typ*, unsigned) const;
  void conv_dense_to_dense(typ*, unsigned) const;  // different leading dim

  // compute minimum diagonal entry of upper triangular matrix
  typ get_min_dentr() const;
};


#endif  // MBLOCK_H
