#ifndef SLL_H
#define SLL_H

#include <assert.h>
#include <iostream>

// Template class for a single concatenated list

template<class T> class sll {

 public:
  // class for a single item in the list
  struct TItem {
    T _data;
    TItem *_next;
    TItem(const T &e, TItem *n = NULL) : _data(e), _next(n) {}
  };
  
  // class for an iterator for the previous list
  class iterator {
    friend class sll<T>;

    TItem *_item;                                // currently selected item

  public:
    // constructor and destructor
    iterator() : _item(NULL) {}
    iterator(TItem *item) : _item(item) {}
    iterator(const iterator &iter) : _item(iter._item) {}
        
    // get status (end-of-list)
    bool eol() const { return (_item == NULL); }
        
    // increment
    iterator &next() { return ++*this; }
        
    // access current element
    T &data() { assert(_item != NULL); return _item->_data; }
    const T &data() const { assert(_item != NULL); return _item->_data; }
    
    // copy
    iterator &operator=(const iterator &i) { _item = i._item; return *this; }

    // access items
    T &operator()() { assert(_item != NULL); return _item->_data; }
    const T &operator()() const { assert(_item != NULL); return _item->_data; }

    // iterate (prefix/postfix)
    iterator &operator++() { if (_item) _item = _item->_next; return *this; }
    iterator operator++(int) { iterator tmp(*this); ++*this; return tmp; }

    // dereference
    T &operator*() { assert(_item != NULL); return _item->_data; }
    T *operator->() { assert(_item != NULL); return &_item->_data; }

    // compare
    bool operator==(const iterator &i) const { return (_item == i._item); }
    bool operator!=(const iterator &i) const { return (_item != i._item); }

  protected:
    // handle case when no item is present
    void no_item () const { std::cerr << "sll::iterator: empty list."
				      << std::endl; }
  };

 protected:
  // first and last element of list
  TItem *_first, *_last;
    
  // size of the list aka number of stored items
  unsigned _size;

 public:
  // constructors and destructors
  sll() : _first(NULL), _last(NULL), _size(0) {}
  sll(const sll<T> &l) : _first(NULL), _last(NULL), _size(0) { copy(l); }

  ~sll() { if (_size>0) remove_all(); }

  // access member variables
  unsigned size() const { return _size; }

  iterator first() { return _first; }
  const iterator first() const { return _first; }

  iterator last() { return _last; }
  const iterator last() const { return _last; }

  // same as above but access real data
  T &first_item() { assert(_first != NULL); return _first->_data; }
  const T &first_item() const { assert(_first != NULL); return _first->_data; }
    
  T &last_item() { assert(_last != NULL); return _last->_data; }
  const T &last_item() const { assert(_last != NULL); return _last->_data; }
    
  // methods for manipulating list

  // add a new object to the begin/end of list
  sll<T> &prepend (const T &elem);
  sll<T> &append  (const T &elem);

  // remove (next) element from list
  void remove(const T &elem);
  void remove(iterator &i) { 
    assert(i._item != NULL);
    remove(i._item->_data);
  }
  void remove_first();

  // return first item and remove it from list
  T behead();
    
  // remove all elements from list
  void remove_all();

  // copy given list into local
  void copy(const sll<T> &list);
    
  // look if element is in list
  bool is_in(const T &elem) const {
    TItem *tmp = _first;

    while (tmp) {
      if (tmp->_data == elem) return true;
      tmp = tmp->_next;
    }

    return false;
  }

  // return size in bytes used by this object
  unsigned long byte_size() const {
    return sizeof(TItem*)+sizeof(TItem*)+sizeof(unsigned)+_size*sizeof(TItem);
  }
  
 protected:
  // handle case when no item is present
  void no_item() const {
    std::cerr << "sll: empty list." << std::endl;
  }
};


template<class T> sll<T> &sll<T>::prepend(const T &elem)
{
  TItem *tmp = new TItem(elem, _first);
    
  _first = tmp;
  _size++;

  if (_last == NULL) _last = tmp;
    
  return *this;
}


template<class T> sll<T> &sll<T>::append(const T &elem)
{
  TItem *tmp = new TItem(elem);
    
  if (_last) _last->_next = tmp;
    
  _last = tmp;
    
  if (!_first) _first = tmp;
    
  _size++;
    
  return *this;
}


template<class T> void sll<T>::remove(const T &elem)
{
  TItem *item = _first;
  TItem *old = NULL;

  while (item != NULL) {
    if (item->_data == elem) {
      if (item == _first)
	_first = item->_next;
      else
	old->_next = item->_next;

      if (item == _last) _last = old;
            
      delete item;
      _size--;

      if ( old == NULL ) 
	item = _first;
      else
	item = old->_next;
    } else {
      old  = item;
      item = item->_next;
    }
  }
}


template<class T> void sll<T>::remove_first()
{
  if (_first) {
    TItem *tmp = _first->_next;
    delete _first;
    if (tmp == NULL) _last = NULL;
    _first = tmp;
    _size--;
  }
}


template<class T> T sll<T>::behead()
{
  assert( _first != NULL );

  T ret = _first->_data;
  TItem *tmp = _first->_next;
  
  delete _first;
  
  if (tmp == NULL) _last = NULL;
  
  _first = tmp;
  _size--;
  
  return ret;
}


template<class T> void sll<T>::remove_all()
{
  if (_size == 0) return;
    
  TItem *next, *tmp = _first;

  while (tmp != NULL) {
    next = tmp->_next;
    delete tmp;
    tmp = next;
  }
    
  _first = _last = NULL;
  _size = 0;
}


template <class T> void sll<T>::copy(const sll<T> &list)
{
  TItem *tmp = list._first;
  TItem *old = NULL;
  TItem *item = list._first;

  remove_all();
    
  while (tmp != NULL) {
    item = new TItem(tmp->_data);

    if (_first == NULL) _first = item;
        
    if (old != NULL) old->_next = item;

    old = item;
    tmp = tmp->_next;
  }

  _last = old;
  _size = list._size;
}
    

#endif   // SLL_H
