#include "LA/BlockMatrix.h"
#include "LA/CG.h";
#include "blas.h"
// CG solves the symmetric positive definite linear
// system Ax=b using the Conjugate Gradient method.
//
// A and C are symmetric H-matrices
//
// The return value indicates convergence within max_iter (input)
// iterations (0), or no convergence within max_iter iterations (1).
//
// Upon successful return, output arguments have the following values:
//  
//      x  --  approximate solution to Ax = b
// nsteps  --  the number of iterations performed before the
//             tolerance was reached
//    eps  --  the residual after the final iteration

template<class IntType>
unsigned CG(BlockMatrix<double, IntType>& A, double* const b, double* const x,
	    double& eps, IntType& nsteps)
{
 


  IntType N = A.NoDofs();
  IntType MIN_NO_STEPS=3;
  double *p, *q, *r, *rhat, rho, rho1=0.0;

  p = new double[4*N];
  q = p + N;
  r = q + N;
  rhat = r + N;
  
  double nrmb = blas::nrm2(N, b);
  if (nrmb==0.0) nrmb = 1.0;

  // r0 = b - Ax0
  blas::copy(N, b, r);
  A.amux(D_MONE, x, r);

  double resid = blas::nrm2(N, r) / nrmb;
  if (resid<eps) {
    eps = resid;
    nsteps = 0;
    delete [] p;
    return 0;
  }

  for (unsigned l=1; l<=nsteps; ++l) {

//#ifndef NDEBUG 
    std::cout << "Step " << l << ", resid=" << resid << std::endl;
    //std::cout << ".";
//#endif

    // r^ = Cr r
	
    blas::copy(N, r, rhat);
	A.precond_apply(rhat);

    // rho = r * r^;
    rho = blas::scpr(N, r, rhat);
    
    if (l>1) {
      double beta = rho / rho1;
      // p = r^ + beta * p
      blas::scal(N, beta, p);
      blas::axpy(N, D_ONE, rhat, p);
    }
    else
      blas::copy(N, rhat, p);

    // q = Ap
    blas::setzero(N, q);
    A.amux(D_ONE, p, q);
	
    // alpha = rho / p*q
    double alpha = rho / blas::scpr(N, p, q);
    
    // x += alpha * p
    blas::axpy(N, alpha, p, x);
    
    // r -= alpha * q
    blas::axpy(N, -alpha, q, r);
    
    resid = blas::nrm2(N, r) / nrmb;

    if (resid<eps && l>MIN_NO_STEPS) {
      eps = resid;
      nsteps = l;
      delete [] p;
      return 0;   
    }

    rho1 = rho;
  }
  
  eps = resid;
  delete [] p;
  return 1;
}


template unsigned CG<unsigned>(BlockMatrix<double, unsigned>& A, double* const b, double* const x,
							   double& eps, unsigned& nsteps);
template unsigned CG<int>(BlockMatrix<double, int>& A, double* const b, double* const x,
						  double& eps, int& nsteps);