#ifndef Shapefunctions_h__
#define Shapefunctions_h__


namespace femlib
{
	//////////////////////////////////////////////////////////////////////////
	// edge functions
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    getEdgeFunctions
	// FullName:  femlib::getEdgeFunctions
	// Access:    public 
	// Returns:   _EXTERN_ double			determinant of the Jacobian at (xi,eta,zeta)
	// Description: Calculates the 6 linear Edge Functions (edge_func) of a curved TET10 at (xi,eta,zeta)
	// Parameter: double xi				
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM					The TET-Node coordinates	
	// Parameter: unsigned * node_labels		The TET-Nodelabels (for global orientation)
	// Parameter: Vec3 * edge_func				The calculated edgefunctions
	//************************************
	_EXTERN_ double getEdgeFunctions(double xi, double eta, double zeta, double* GEOM, unsigned* node_labels,Vec3* edge_func);
	//************************************
	// Method:    getEdgeFunctions
	// FullName:  femlib::getEdgeFunctions_flat
	// Access:    public 
	// Returns:   _EXTERN_ double			determinant of the Jacobian at (xi,eta,zeta)
	// Description: Calculates the 6 linear, flat Edge Functions (edge_func) of a flat TET4 at (xi,eta,zeta)
	// Parameter: double xi				
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM					The TET4-Node coordinates	
	// Parameter: unsigned * node_labels		The TET4-Nodelabels (for global orientation)
	// Parameter: Vec3 * edge_func				The calculated edgefunctions
	//************************************
	_EXTERN_ double getEdgeFunctions_flat(double xi, double eta, double zeta, double* GEOM, unsigned* node_labels,Vec3* edge_func);

	//! Returns unoriented Edge Functions for point (xi,eta,zeta) (just used in getEdgeFunctions)
	//************************************
	// Method:    getUnorientedEdgeFunctions
	// FullName:  femlib::getUnorientedEdgeFunctions
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: Calculates the 6 linear Edge Functions (edge_func) of a curved TET10 at (xi,eta,zeta) without global orientation
	// Parameter: double xi				
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM				The TET-Node coordinates	
	// Parameter: Vec3 * edge_func			The calculates edgefunctions
	//************************************
	_EXTERN_ double getUnorientedEdgeFunctions(double xi, double eta, double zeta, double* GEOM, Vec3* edge_func);


	//! Returns unoriented Edge Functions for point (xi,eta,zeta) (just used in getEdgeFunctions)
	//************************************
	// Method:    getUnorientedEdgeFunctions
	// FullName:  femlib::getUnorientedEdgeFunctions_flat
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: Calculates the 6 linear Edge Functions (edge_func) of a flat TET4 at (xi,eta,zeta) without global orientation
	// Parameter: double xi				
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM				The TET4-Node coordinates	
	// Parameter: Vec3 * edge_func			The calculates edgefunctions
	//************************************
	_EXTERN_ double getUnorientedEdgeFunctions_flat(double xi, double eta, double zeta, double* GEOM, Vec3* edge_func);

	//************************************
	// Method:    getEdgeFunctionsAtQP
	// FullName:  femlib::getEdgeFunctionsAtQP
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Calculate the edgefunctions (edge_func) at the QPs (of the QTET4 routine)
	// Parameter: double * GEOM				The TET-Node coordinates	
	// Parameter: unsigned * node_labels	The TET-Nodelabels (for global orientation)
	// Parameter: Vec3 *  & edge_func		The resulting edgefunctions. 
	//										Stored with: 
	//						for(qp_i=0;qp_i<4;qp_i++)
	//							for(edge_func_i=0;edge_func_i<6;edge_func_i++)
	//								edge_func[ge(edge_func_i,qp_i,6)]=EDGEFUNC[edge_func_i];		
	//
	// Parameter: double * dets		dets[qp_i]=fabs(det);	
	//************************************
	_EXTERN_ void getEdgeFunctionsAtQP(double* GEOM,unsigned* node_labels, Vec3* &edge_func, double* dets);
	//************************************
	// Method:    getEdgeFunctionsAtQP
	// FullName:  femlib::getEdgeFunctionsAtQP_flat
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Calculate the flat edgefunctions (edge_func) at the QPs (of the QTET4 routine)
	// Parameter: double * GEOM				The TET4-Node coordinates	
	// Parameter: unsigned * node_labels	The TET4-Nodelabels (for global orientation)
	// Parameter: Vec3 *  & edge_func		The resulting edgefunctions. 
	//										Stored with: 
	//						for(qp_i=0;qp_i<4;qp_i++)
	//							for(edge_func_i=0;edge_func_i<6;edge_func_i++)
	//								edge_func[ge(edge_func_i,qp_i,6)]=EDGEFUNC[edge_func_i];		
	//
	// Parameter: double * dets		dets[qp_i]=fabs(det);	
	//************************************
	_EXTERN_ void getEdgeFunctionsAtQP_flat(double* GEOM,unsigned* node_labels, Vec3* &edge_func, double* dets);

	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    N4
	// FullName:  femlib::N4
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Calculate the 4 linear main nodal shapefunctions of a curved TET10 
	//				(interpolated from the tet10 shapefunctions)
	// Parameter: double xi
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double *  & shape_fun		
	//************************************
	_EXTERN_ void N4(const double xi, const double eta, const double zeta, double* &shape_fun);
	//////////////////////////////////////////////////////////////////////////
} // end femlib


#endif // Shapefunctions_h__